;;;; util.huffman -- generalized huffman codes

;;; Created:    <2002-05-10 11:12:37 foof>
;;; Time-stamp: <2003-04-11 15:26:38 foof>
;;; Author:     Alex Shinn <foof@synthcode.com>

(define-module util.huffman
  (use srfi-1)    ;; list library
  (use util.sort) ;; merge procedure
  (use gauche.collection)  ;; generalize map and for-each
  (export make-huffman-code make-huffman-stats
          huffman-encode huffman-decode
          huffman-decode-string huffman-decode-vector))
(select-module util.huffman)

;; weighted tree... leaf is (value . weight), node is (left right . weight)

(define (huff-leaf? obj) (and (pair? obj) (not (pair? (cdr obj)))))
(define huff-weight (lambda (x) (if (huff-leaf? x) (cdr x) (cddr x))))
(define huff-value car)
(define huff-left car)
(define huff-right cadr)

(define (huff-node a b weight) (cons a (cons b weight)))

(define (hash->alist table)
  (hash-table-map table cons))

(define (make-huffman-stats obj)
  (let ((table (make-hash-table 'equal?)))
    (for-each
     (lambda (x)
       (hash-table-put! table x (+ 1 (hash-table-get table x 0))))
     obj)
    (hash->alist table)))

(define (merge-stats a b)
  (let ((sum (+ (huff-weight a) (huff-weight b))))
    (huff-node a b sum)))

;; collect stats into a weighted tree format
(define (make-huffman-tree lst)
  (if (null? (cdr lst))
    (car lst)
    (make-huffman-tree (merge (list (merge-stats (car lst) (cadr lst)))
                              (cddr lst)
                              (lambda (a b) (< (huff-weight a)
                                               (huff-weight b)))))))

;; build a hash table mapping values to huffman codes
(define (make-huffman-table tree)
  (define table (make-hash-table 'equal?))
  (define path->huff-code reverse)
  (define (helper node path)
    (if (null? node)
      #f
      (cond ((huff-leaf? node)
             ;;(format #t "leaf node: ~S\n" node)
             (hash-table-put! table (huff-value node)
                              (path->huff-code path) )
             )
            (else
             ;;(format #t "non-leaf: ~S\n" node)
             (helper (huff-left node) (cons 0 path))
             (helper (huff-right node) (cons 1 path)) ))))
  (helper tree '())
  table)

(define (print-path p)
  (cond ((not (null? p))
         (display (car p))
         (print-path (cdr p)))))

(define (path-prefix? p1 p2)
  (cond ((null? p1)
         #t)
        ((null? p2)
         #f)
        ((eq? (car p1) (car p2))
         (path-prefix? (cdr p1) (cdr p2)))
        (else
         #f)))

(define (print-huffman-code table)
  (hash-table-for-each
   table
   (lambda (k v)
     (write k)
     (display ": ")
     (print-path v)
     (newline))))

(define (validate-huffman-code table)
  (define (bad-paths p1 p2)
    (print-path p1)
    (display " is a prefix of ")
    (print-path p2)
    (newline))
  (let outer ((codes (hash->alist table)))
    (if (not (null? codes))
      (let ((path1 (cdar codes)))
        (let inner ((rest (cdr codes)))
          (if (not (null? rest))
            (let ((path2 (cdar rest)))
              (if (path-prefix? path1 path2)
                (bad-paths path1 path2))
              (if (path-prefix? path2 path1)
                (bad-paths path2 path1))
              (inner (cdr rest))))))
      (outer (cdr codes)))))

;; zero pad
(define (bitlist->string lst)
  (with-output-to-string
    (lambda ()
      (let* ((len (length lst))
             (rem (modulo len 8)))
        (if (not (= rem 0))
          (set! lst (append lst (make-list (- 8 rem) 0))))
        (let loop ((l lst))
          (cond
            ((null? l)
             #f)
            (else
             (let build-char ((i 0) (c 0) (bit l))
               (cond
                 ((= i 8)
                  (write-byte c)
                  (loop bit))
                 (else
                  (build-char (+ i 1) (+ (* c 2) (car bit)) (cdr bit))))))))))))

(define (char->bitlist x)
  (list
   (min 1 (logand x #b10000000))
   (min 1 (logand x #b01000000))
   (min 1 (logand x #b00100000))
   (min 1 (logand x #b00010000))
   (min 1 (logand x #b00001000))
   (min 1 (logand x #b00000100))
   (min 1 (logand x #b00000010))
          (logand x #b00000001)
   ))

(define (string->bitlist str)
  (with-input-from-string str
    (lambda ()
      (let loop ((res '()))
        (let ((c (read-byte)))
          (if (eof-object? c)
            (concatenate (reverse res))
            (loop (cons (char->bitlist c) res))))))))

(define (x->bitlist x)
  (cond ((string? x) (string->bitlist x))
        ((vector? x) (vector->list x))
        ((pair? x)   x)
        (else
         (error "can't convert to bitlist " x))))

(define (make-huffman-code stats . opt-sentinel)
  (let* ((sentinel (if (pair? opt-sentinel) (car opt-sentinel) (gensym)))
         (min-stat (apply min (cons 1 (map cdr stats))))
         (tree (make-huffman-tree (cons (cons sentinel min-stat) stats)))
         (table (make-huffman-table tree))
         (end-code (hash-table-get table sentinel)))
    ;;(format #t "stats: ~S\ntree: ~S\ntable: ~S\n" stats tree table)
    ;;(print-huffman-code table)
    (validate-huffman-code table)
    (letrec ((encode-one
              (lambda (obj) (hash-table-get table obj)))
             (encode
              (lambda (seq)
                (let ((res '()))
                  (for-each
                   (lambda (x) (push! res (encode-one x)))
                   seq)
                  (bitlist->string (concatenate (reverse (cons end-code res)))))))
             (decode-path
              (lambda (path node)
                (cond ((null? path)
                       (if (huff-leaf? node)
                         (huff-value node)
                         (error "incomplete encoding")))
                      ((huff-leaf? node)
                       (error "invalid encoding"))
                      ((= (car path) 0)
                       (decode-path (cdr path) (huff-left node)))
                      (else
                       (decode-path (cdr path) (huff-right node))))))
             (decode
              (lambda (encoding)
                (let ((res '()))
                  (let loop ((code (x->bitlist encoding))
                             (node tree))
                    (cond ((huff-leaf? node)
                           (let ((v (huff-value node)))
                             (cond ((eq? v sentinel)
                                    #f) ;; stop at the sentinel
                                   (else
                                    (set! res (cons (huff-value node) res))
                                    (loop code tree)))))
                          ((null? code)
                           #f)
                          ((= (car code) 0)
                           (loop (cdr code) (huff-left node)))
                          (else
                           (loop (cdr code) (huff-right node)))))
                  (reverse res))))
             )
      (lambda (command . args)
        (case command
          ;; basic interface
          ((encode)
           (encode (car args)))
          ((decode decode-list)
           (decode (car args)))
          ((decode-string)
           (list->string (decode (car args))))
          ((decode-vector)
           (list->vector (decode (car args))))
          ((encoder) encode)
          ((decoder list-decoder) decode)
          ((string-decoder)
           (lambda (x) (list->string (decode x))))
          ((vector-decoder)
           (lambda (x) (list->vector (decode x))))
          ;; introspection
          ((get-sentinel)   sentinel)
          ((get-table)      table)
          ((get-tree)       tree)
          ((set-sentinel!)  (set! sentinel (car args)))
          ((set-table!)     (set! table (car args)))
          ((set-tree!)      (set! tree (car args)))
          (else
           (error "unknown command " command)))))))

;; convenience procedures

(define (huffman-encode seq . opt-stats)
  (let* ((stats (if (pair? opt-stats)
                  (car opt-stats)
                  (make-huffman-stats seq)))
         (huff (make-huffman-code stats)))
    (list (huff 'encode seq) huff)))

(define (make-huffman-decoder command)
  (lambda (seq stats . opt-sentinel)
    (let ((huff (apply make-huffman-code (append (list stats) opt-sentinel))))
      (list (huff command seq) huff))))

(define huffman-decode (make-huffman-decoder 'decode))
(define huffman-decode-string (make-huffman-decoder 'decode-string))
(define huffman-decode-vector (make-huffman-decoder 'decode-vector))

(provide "util/huffman")

